{$APPTYPE CONSOLE}

{$DEFINE UseASM}
{$DEFINE UseSHL}

PROGRAM
  MPEGWalk;

USES
  Windows, SysUtils, HexFuncs;

TYPE
  TimeCode = RECORD H, M, S, F: BYTE END;

CONST
  MaxBufSize = $4000;
  RotChar: ARRAY[0..3] OF CHAR = '-\|/';
  FrameRate: ARRAY[1..8] OF DOUBLE = (24000/1001, 24.0, 25.0, 30000/1001, 30.0, 50.0, 60000/1001, 60.0);
  FrameType: ARRAY[0..7] OF CHAR = 'xIPBDrrr';
  ShiftMask = $00FFFFFF;

VAR
  CON, DAT: TEXT;
  F: FILE;
  B, FRCode, GOPPics, FType, StreamID, SubStreamID: BYTE;
  FRate, PicRate, GOPRate, AllRate: DOUBLE;
  I: INTEGER;
  BufPos, BufSize, FGOPPos, PacketLen: WORD;
  L, PL, BR, PicCount, SliceBits, GOPBits: CARDINAL;
  AllBits: COMP;
  Buf: ARRAY[0..Pred(MaxBufSize)] OF BYTE;
  verbose, inSlice, inSystem,
  _PTS, _DTS, _ESCR, _ESRate, _Trick, _Copy, _CRC, _Ext,
  __PESPriv, __PHF, __PPSC, __PSTD, __PESX: BOOLEAN;
  TC: TimeCode;
  Param, FPath, FDir, FName, FExt: STRING;
  GOPStruc: PChar;

FUNCTION FormatTC: STRING;
  VAR S, T: STRING;
  BEGIN
    S := '';
    Str(TC.H, T);
    IF TC.H < 10 THEN S := S + '0';
    S := S + T + ':';
    Str(TC.M, T);
    IF TC.M < 10 THEN S := S + '0';
    S := S + T + ':';
    Str(TC.S, T);
    IF TC.S < 10 THEN S := S + '0';
    S := S + T + '.'; 
    Str(TC.F, T); 
    IF TC.F < 10 THEN S := S + '0'; 
    S := S + T; 
    FormatTC := S
  END; 

PROCEDURE EndOfPicture; 
  BEGIN
    inSlice := false; 
    GOPBits := GOPBits + SliceBits; 
    IF SliceBits > 0 THEN BEGIN
      IF verbose THEN WriteLn('{GOP += ', SliceBits, ' => ', GOPBits, '}'); 
      PicRate := FRate*SliceBits/1e6; 
      Write('> ', SliceBits, ' bits in previous picture = '); 
      WriteLn(PicRate:4:2, ' Mbps/Picture')
    END; 
    SliceBits := 0
  END; 

PROCEDURE EndOfGOP; 
  BEGIN
    EndOfPicture;
    IF GOPPics > 0 THEN BEGIN
      GOPRate := FRate/GOPPics*GOPBits/1e6;
      Write('>> ', GOPBits, ' bits in ',GOPPics, ' pictures in previous GOP = ');
      WriteLn(GOPRate:4:2, ' Mbps/GOP');
      WriteLn('(GOP structure = ', GOPStruc, ')')
    END;
    GOPPics := 0;
    GOPBits := 0;
    FillChar(GOPStruc^, SizeOf(GOPStruc^), #0);
  END;

{$IFDEF UseASM}
PROCEDURE ShiftAddNext(B: BYTE); ASSEMBLER;
  ASM
    MOV BL, B
    MOV EAX, DWORD Ptr L
    MOV DWORD Ptr PL, EAX
    {$IFDEF UseSHL}
    SHL EAX, 8
    {$ELSE}
    XOR EDX, EDX
    MOV EDX, $0100
    MUL EDX
    {$ENDIF}
    MOV AL, BL
    MOV DWORD Ptr L, EAX
  END;
{$ELSE}
PROCEDURE ShiftAddNext(B: BYTE);
  BEGIN
    PL := L;
    L := (PL SHL 8) OR B
  END;
{$ENDIF}

FUNCTION NextByte: BYTE; 
  BEGIN
    Result := 0;
    Inc(BufPos);
    IF inSlice AND NOT inSystem
    THEN BEGIN
      Inc(SliceBits, 8);
      AllBits := AllBits + 8
    END;
    IF PacketLen > 0 THEN Dec(PacketLen);
    IF BufSize = 0 THEN Exit;
    IF BufPos >= BufSize THEN BEGIN
      BufPos := BufPos MOD BufSize;
      BlockRead(F, Buf, BufSize, I);
      BufSize := I;
    END;
    Result := Buf[BufPos];
    ShiftAddNext(Buf[BufPos])
  END; 

PROCEDURE NextNBytes(N: WORD); 
  VAR Count: WORD; 
  BEGIN FOR Count := 1 TO N DO B := NextByte END; 

BEGIN
  Assign(CON, 'CON');
  Rewrite(CON);
  WriteLn(CON);
  WriteLn(CON, ^I'MPEG Walk');
  WriteLn(CON);
  L := $FFFFFFFF;
  GetMem(GOPStruc, 64);
  FillChar(GOPStruc^, 64, #0);
  FileMode := 0;
  BufSize := MaxBufSize;
  BufPos := Pred(BufSize);
  PicCount := 0;
  SliceBits := 0;
  GOPBits := 0;
  GOPPics := 0;
  GOPRate := 0.0;
  AllBits := 0;
  AllRate := 0.0;
  PacketLen := 0;
  verbose := (ParamCount > 1); 
  IF ParamCount > 0 THEN BEGIN
    Param := ParamStr(1); 
    Assign(F, Param); 
    Reset(F, 1); 
    I := IOResult; 
    IF I = 0 THEN BEGIN
      FPath := ExpandFileName(Param);
      FDir := ExtractFileDir(FPath);
      FName := ExtractFileName(FPath);
      FExt := ExtractFileExt(FPath);
      Assign(DAT, ChangeFileExt(FPath, '.LST'));
      Rewrite(DAT); 
      WriteLn(DAT, Param);
      WriteLn(DAT, 'Timecode'^I'Bitrate'); 
      REPEAT
        B := NextByte; 
        IF (PL AND ShiftMask) = $000001 THEN BEGIN
          IF verbose THEN WriteLn('---'); 
          inSystem := (B > $AF); 
          CASE B OF
            $00: BEGIN {Picture Start Code}
                   EndOfPicture; 
                   Inc(GOPPics); 
                   Inc(PicCount); 
                   Write(CON, RotChar[PicCount AND 3], ' ', PicCount, ^M); 
                   Write('[', HexL(L), ']', ' = Picture start'); 
                   NextNBytes(2); 
                   FType := (B SHR 3) AND 7; 
                   FGOPPos := (L AND $FFC0) SHR 6;
                   GOPStruc[FGopPos] := FrameType[FType];
                   WriteLn(' (', FrameType[FType], ' frame at GOP position ', FGOPPos, ')'); 
                   IF NOT (FType IN [1..4]) THEN BEGIN
                     WriteLn('>>ERROR<< - Unsupported Picture Type ', FType); 
                     {Halt}
                   END
                 END; 
            $01..$AF: BEGIN {Slice Start Code}
                        inSlice := TRUE; 
                        IF verbose THEN WriteLn('[', HexL(L), ']', ' = Slice (row ', 16*WORD(Pred(B)), ')')
                      END; 
            $B0, $B1: BEGIN {RESERVED}
                        WriteLn('[', HexL(L), ']', ' = >RESERVED<')
                      END; 
            $B2: BEGIN {User Data Start Code}
                   WriteLn('[', HexL(L), ']', ' = User Data Start code')
                 END; 
            $B3: BEGIN {Sequence Header Code}
                   EndOfGOP; 
                   WriteLn('[', HexL(L), ']', ' = Sequence Header code'); 
                   Write('(Framerate: '); 
                   NextNBytes(4); 
                   FRCode := B AND $0F; 
                   FRate := FrameRate[FRCode]; 
                   Write(FRate:0:3, ' fps; max. Bitrate: '); 
                   NextNBytes(4); 
                   BR := (PL AND ShiftMask) SHR 6; 
                   WriteLn(BR/2500:4:2, ' Mbps)')
                 END; 
            $B4: BEGIN {Sequence Error Code}
                   EndOfGOP; 
                   WriteLn('[', HexL(L), ']', ' = Sequence Error code')
                 END; 
            $B5: BEGIN {Extension Start Code}
                   Write('[', HexL(L), ']', ' = Extension Start code '); 
                   B := NextByte; 
                   Write(HexDigit[B SHR 4], ' - '); 
                   CASE B OF
                     $10..$1F: WriteLn('Sequence Extension'); 
                     $20..$2F: WriteLn('Sequence Display Extension'); 
                     $30..$3F: WriteLn('Quant Matrix Extension'); 
                     $50..$5F: WriteLn('Sequence Scalable Extension'); 
                     $70..$7F: WriteLn('Picture Display Extension'); 
                     $80..$8F: WriteLn('Picture Coding Extension'); 
                     $90..$9F: WriteLn('Picture Spatial Scalable Extension'); 
                     $A0..$AF: WriteLn('Picture Temporal Scalable Extension'); 
                     ELSE WriteLn('>RESERVED<'); 
                   END
                 END; 
            $B6: BEGIN {RESERVED}
                   WriteLn('[', HexL(L), ']', ' = >RESERVED<')
                 END; 
            $B7: BEGIN {Sequence End Code}
                   EndOfGOP; 
                   WriteLn('[', HexL(L), ']', ' = Sequence End code'); 
                   AllRate := (AllBits/PicCount)*FRate/1e6; 
                   Write('>>> ', AllBits:1:0, ' bits in ', PicCount:0, ' pictures in the movie = '); 
                   WriteLn(AllRate:4:2, ' Mbps/movie')
                 END; 
            $B8: BEGIN {Group Start Code}
                   IF GOPRate > 0 THEN BEGIN
                     WriteLn(DAT, FormatTC, ^I, GOPRate:0:3); 
                     Flush(DAT)
                   END; 
                   EndOfGOP; 
                   WriteLn('[', HexL(L), ']', ' = Group Start code'); 
                   NextNBytes(4); 
                   PL := (L SHR 7); 
                   TC.F := (PL AND $3F); 
                   PL := (PL SHR 6); 
                   TC.S := (PL AND $3F); 
                   PL := (PL SHR 7); 
                   TC.M := (PL AND $3F); 
                   PL := (PL SHR 6); 
                   TC.H := (PL AND $1F); 
                   WriteLn('(Timecode = ', FormatTC, ')'); 
                 END; 
            $B9..$FF: BEGIN { (System Start codes) }
                        Write('[', HexL(L), ']', ' = {SYSTEM START CODE} - '); 
                        StreamID := B; 
                        CASE StreamID OF
                          $B9:
                                WriteLn('Program End code'); 
                          $BA:
                                BEGIN
                                  WriteLn('Pack Start code'); 
                                  NextNBytes(10); 
                                  NextNBytes(B AND 7)
                                END; 
                          $BB:
                                WriteLn('System Header Start code'); 
                          $BC:
                                WriteLn('Program Stream Map'); 
                          $BD:
                                WriteLn('Private Stream 1'); 
                          $BE:
                                WriteLn('Padding Stream'); 
                          $BF:
                                WriteLn('Private Stream 2'); 
                          $C0..$DF:
                                     WriteLn('Audio Stream ', B AND $3F); 
                          $E0..$EF:
                                     WriteLn('Video Stream ', B AND $0F); 
                          $F0:
                                WriteLn('ECM Stream'); 
                          $F1:
                                WriteLn('EMM Stream'); 
                          $F2:
                                WriteLn('DSM-CC Stream'); 
                          $F3:
                                WriteLn('ISO/IEC-13522 Stream'); 
                          $F4..$FE:
                                     WriteLn('>RESERVED< Data Stream'); 
                          $FF:
                                WriteLn('Program Stream Directory'); 
                          ELSE WriteLn('[', HexB(B), ']'); 
                        END; 
                        IF StreamID > $BA THEN BEGIN
                          NextNBytes(2); 
                          PacketLen := WORD(L); 
                          WriteLn('(Packet length: ', PacketLen, ' bytes)'); 
                          CASE StreamID OF
                            $BE:
                                  BEGIN
                                    WriteLn('(Skipping ', PacketLen, ' stuffing bytes)'); 
                                    NextNBytes(PacketLen)
                                  END; 
                            $BB, $BC, $BF, $F0, $F1, $FF:
                                                           BEGIN
                                                             WriteLn('(Skipping ', PacketLen, ' data bytes)'); 
                                                             NextNBytes(PacketLen)
                                                           END; 
                            ELSE BEGIN
                              WriteLn('(Reading sync info)'); 
                              NextNBytes(3); 
                              _PTS := (L AND $8000) > 0; 
                              _DTS := (L AND $4000) > 0; 
                              _ESCR := (L AND $2000) > 0; 
                              _ESRate := (L AND $1000) > 0; 
                              _Trick := (L AND $0800) > 0; 
                              _Copy := (L AND $0400) > 0; 
                              _CRC := (L AND $0200) > 0; 
                              _Ext := (L AND $0100) > 0; 
                              IF _PTS THEN NextNBytes(5); 
                              IF _DTS THEN NextNBytes(5); 
                              IF _ESCR THEN NextNBytes(6); 
                              IF _ESRate THEN NextNBytes(3); 
                              IF _Trick THEN NextNBytes(1); 
                              IF _Copy THEN NextNBytes(1); 
                              IF _CRC THEN NextNBytes(2); 
                              IF _Ext THEN BEGIN
                                NextNBytes(1); 
                                __PESPriv := (B AND $80) > 0; 
                                __PHF := (B AND $40) > 0; 
                                __PPSC := (B AND $20) > 0; 
                                __PSTD := (B AND $10) > 0; 
                                __PESX := (B AND $01) > 0; 
                                IF __PESPriv THEN NextNBytes(7); 
                                IF __PHF THEN NextNBytes(1); 
                                IF __PPSC THEN NextNBytes(2); 
                                IF __PSTD THEN NextNBytes(2); 
                                IF __PESX THEN BEGIN
                                  NextNBytes(1); 
                                  NextNBytes(B AND $7F)
                                END
                              END; 
                              CASE StreamID OF
                                $E0..$EF:
                                           inSystem := FALSE; 
                                $BD:
                                      BEGIN
                                        IF PacketLen > 0 THEN REPEAT
                                          B := NextByte
                                        UNTIL B <> $FF; 
                                        SubStreamID := B; 
                                        Write('- Substream ', HexB(SubStreamID)); 
                                        CASE SubStreamID OF
                                          $20..$3F:
                                                     Write(': Subpicture; '); 
                                          $80..$87:
                                                     Write(': AC-3 Audio; '); 
                                          $88..$8F:
                                                     Write(': dts Audio; '); 
                                          $A0..$A7:
                                                     Write(': PCM Audio; '); 
                                          ELSE WriteLn
                                        END; 
                                        NextNBytes(1); 
                                        IF SubStreamID IN [$80..$8F, $A0..$A7]
                                        THEN Write(B, ' sync words, 1st at offset ')
                                        ELSE Write(HexB(B), ' '); 
                                        NextNBytes(2); 
                                        IF SubStreamID IN [$80..$8F, $A0..$A7]
                                        THEN WriteLn(WORD(L))
                                        ELSE WriteLn(HexB(Hi(WORD(L))), ' ', HexB(Lo(WORD(L))));
                                        WriteLn('(Skipping ', PacketLen, ' data bytes)'); 
                                        NextNBytes(PacketLen)
                                      END
                                      ELSE BEGIN
                                        WriteLn('(Skipping ', PacketLen, ' bytes)'); 
                                        NextNBytes(PacketLen)
                                      END
                              END
                            END
                          END
                        END
                      END; 
          END
        END
      UNTIL BufSize = 0;
      WriteLn('____________________');
      WriteLn;
      AllRate := (AllBits/PicCount)*FRate/1e6;
      Write('>>> ', AllBits:1:0, ' bits in ', PicCount:0, ' pictures in the movie = ');
      WriteLn(AllRate:4:2, ' Mbps/movie');
      Close(F); 
      Close(DAT); 
      WriteLn(CON, ^I'O.K.')
    END; 
    Close(CON)
  END
  ELSE WriteLn('Syntax: MPEGWALK MPEGFILE [/Verbose]')
END.
